"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _bn = _interopRequireDefault(require("bn.js"));

var _bignumber = _interopRequireDefault(require("bignumber.js"));

var _errors = require("./errors");

var _utils = _interopRequireDefault(require("./utils"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var add0x = function add0x(value) {
  return "0x".concat(value);
};

var strip0x = function strip0x(value) {
  return value.startsWith('0x') ? value.substring(2) : value;
}; // eslint-disable-next-line


var UINT160_BUFFER_BYTES = 20;

var bufferToUInt160 = function bufferToUInt160(value) {
  if (value.length !== UINT160_BUFFER_BYTES) {
    throw new _errors.InvalidFormatError();
  }

  return value;
};

var uInt160ToHex = function uInt160ToHex(value) {
  return add0x(_utils.default.reverse(value).toString('hex'));
};

var hexToUInt160 = function hexToUInt160(value) {
  return typeof value === 'string' ? bufferToUInt160(_utils.default.reverse(Buffer.from(strip0x(value), 'hex'))) : value;
};

var uInt160ToBuffer = function uInt160ToBuffer(value) {
  return typeof value === 'string' ? hexToUInt160(value) : value;
};

var uInt160ToString = function uInt160ToString(value) {
  return typeof value === 'string' ? value : uInt160ToHex(value);
};

var stringToUInt160 = function stringToUInt160(value) {
  return hexToUInt160(value);
};

var uInt160Equal = function uInt160Equal(a, b) {
  return a.equals(b);
};

var isUInt160 = function isUInt160(value) {
  return value instanceof Buffer && value.length === UINT160_BUFFER_BYTES;
};

var asUInt160 = function asUInt160(value) {
  if (value instanceof Buffer) {
    return bufferToUInt160(value);
  }

  throw new _errors.InvalidFormatError();
};

var ZERO_UINT160 = Buffer.alloc(20, 0); // eslint-disable-next-line

var UINT256_BUFFER_BYTES = 32;
var ZERO_UINT256 = Buffer.alloc(UINT256_BUFFER_BYTES, 0);

var bufferToUInt256 = function bufferToUInt256(value) {
  if (value.length !== UINT256_BUFFER_BYTES) {
    throw new _errors.InvalidFormatError();
  }

  return value;
};

var uInt256ToHex = function uInt256ToHex(value) {
  return add0x(_utils.default.reverse(value).toString('hex'));
};

var hexToUInt256 = function hexToUInt256(value) {
  return typeof value === 'string' ? bufferToUInt256(_utils.default.reverse(Buffer.from(strip0x(value), 'hex'))) : value;
};

var uInt256ToBuffer = function uInt256ToBuffer(value) {
  return typeof value === 'string' ? hexToUInt256(value) : value;
};

var uInt256ToString = function uInt256ToString(value) {
  return typeof value === 'string' ? value : uInt256ToHex(value);
};

var stringToUInt256 = function stringToUInt256(value) {
  return hexToUInt256(value);
};

var uInt256Equal = function uInt256Equal(a, b) {
  return a.equals(b);
};

var toUInt32LE = function toUInt32LE(bytes) {
  return new _bn.default(uInt256ToBuffer(bytes).slice(0, 4), 'le').toNumber();
}; // eslint-disable-next-line


// Encoded compressed ECPoint
var ECPOINT_BUFFER_BYTES = 33;
var ECPOINT_INFINITY_BYTE = 0x00;
var ECPOINT_INFINITY = Buffer.from([ECPOINT_INFINITY_BYTE]);

var ecPointToHex = function ecPointToHex(value) {
  return typeof value === 'string' ? value : value.toString('hex');
};

var hexToECPoint = function hexToECPoint(value) {
  return typeof value === 'string' ? Buffer.from(value, 'hex') : value;
};

var ecPointToBuffer = function ecPointToBuffer(value) {
  return typeof value === 'string' ? hexToECPoint(value) : value;
};

var bufferToECPoint = function bufferToECPoint(value) {
  if (value.length !== ECPOINT_BUFFER_BYTES) {
    if (value.length === 1 && value.equals(ECPOINT_INFINITY)) {
      return value;
    }

    throw new _errors.InvalidFormatError();
  }

  return value;
};

var ecPointToString = function ecPointToString(value) {
  return typeof value === 'string' ? value : ecPointToHex(value);
};

var stringToECPoint = function stringToECPoint(value) {
  return hexToECPoint(value);
};

var ecPointEqual = function ecPointEqual(a, b) {
  return a.equals(b);
};

var ecPointCompare = function ecPointCompare(a, b) {
  var aHex = ecPointToHex(a);
  var bHex = ecPointToHex(b);

  if (aHex < bHex) {
    return -1;
  } else if (aHex > bHex) {
    return 1;
  }

  return 0;
};

var ecPointIsInfinity = function ecPointIsInfinity(value) {
  return value.equals(ECPOINT_INFINITY);
}; // eslint-disable-next-line


var PRIVATE_KEY_BUFFER_BYTES = 32;

var privateKeyToHex = function privateKeyToHex(value) {
  return typeof value === 'string' ? value : value.toString('hex');
};

var hexToPrivateKey = function hexToPrivateKey(value) {
  return typeof value === 'string' ? Buffer.from(value, 'hex') : value;
};

var privateKeyToBuffer = function privateKeyToBuffer(value) {
  return typeof value === 'string' ? hexToECPoint(value) : value;
};

var bufferToPrivateKey = function bufferToPrivateKey(value) {
  if (value.length !== PRIVATE_KEY_BUFFER_BYTES) {
    throw new _errors.InvalidFormatError();
  }

  return value;
};

var privateKeyToString = function privateKeyToString(value) {
  return typeof value === 'string' ? value : ecPointToHex(value);
};

var stringToPrivateKey = function stringToPrivateKey(value) {
  return hexToECPoint(value);
};

var D = new _bn.default(100000000);
var DBigNumber = new _bignumber.default(D.toString(10));

var fixed8FromDecimal = function fixed8FromDecimal(value) {
  if (typeof value === 'number') {
    return new _bn.default(value).mul(D);
  }

  if (value instanceof _bn.default) {
    return value.mul(D);
  }

  var valueBigNumber = value;

  if (typeof value === 'string') {
    valueBigNumber = new _bignumber.default(value);
  } // $FlowFixMe


  return new _bn.default(valueBigNumber.times(DBigNumber).toString(), 10);
};

var fixed8ToDecimal = function fixed8ToDecimal(bn) {
  return new _bignumber.default(bn.toString(10)).div(DBigNumber);
};

var NEGATIVE_SATOSHI_FIXED8 = new _bn.default(-1);
var ONE_HUNDRED_FIXED8 = fixed8FromDecimal(100);
var _default = {
  D: D,
  NEO_ADDRESS_VERSION: 23,
  NEO_PRIVATE_KEY_VERSION: 0x80,
  ECPOINT_BUFFER_BYTES: ECPOINT_BUFFER_BYTES,
  ECPOINT_INFINITY: ECPOINT_INFINITY,
  ECPOINT_INFINITY_BYTE: ECPOINT_INFINITY_BYTE,
  PRIVATE_KEY_BUFFER_BYTES: PRIVATE_KEY_BUFFER_BYTES,
  UINT160_BUFFER_BYTES: UINT160_BUFFER_BYTES,
  UINT256_BUFFER_BYTES: UINT256_BUFFER_BYTES,
  ZERO_UINT160: ZERO_UINT160,
  ZERO_UINT256: ZERO_UINT256,
  NEGATIVE_SATOSHI_FIXED8: NEGATIVE_SATOSHI_FIXED8,
  ONE_HUNDRED_FIXED8: ONE_HUNDRED_FIXED8,
  uInt160ToBuffer: uInt160ToBuffer,
  bufferToUInt160: bufferToUInt160,
  uInt160ToHex: uInt160ToHex,
  hexToUInt160: hexToUInt160,
  uInt160ToString: uInt160ToString,
  stringToUInt160: stringToUInt160,
  uInt160Equal: uInt160Equal,
  isUInt160: isUInt160,
  asUInt160: asUInt160,
  uInt256ToBuffer: uInt256ToBuffer,
  bufferToUInt256: bufferToUInt256,
  uInt256ToHex: uInt256ToHex,
  hexToUInt256: hexToUInt256,
  uInt256ToString: uInt256ToString,
  stringToUInt256: stringToUInt256,
  uInt256Equal: uInt256Equal,
  toUInt32LE: toUInt32LE,
  ecPointToBuffer: ecPointToBuffer,
  bufferToECPoint: bufferToECPoint,
  ecPointToHex: ecPointToHex,
  hexToECPoint: hexToECPoint,
  ecPointToString: ecPointToString,
  ecPointCompare: ecPointCompare,
  stringToECPoint: stringToECPoint,
  ecPointEqual: ecPointEqual,
  ecPointIsInfinity: ecPointIsInfinity,
  privateKeyToHex: privateKeyToHex,
  hexToPrivateKey: hexToPrivateKey,
  privateKeyToBuffer: privateKeyToBuffer,
  bufferToPrivateKey: bufferToPrivateKey,
  privateKeyToString: privateKeyToString,
  stringToPrivateKey: stringToPrivateKey,
  fixed8FromDecimal: fixed8FromDecimal,
  fixed8ToDecimal: fixed8ToDecimal
};
exports.default = _default;